package ru.dbotthepony.mc.otm.item.weapon

import net.minecraft.sounds.SoundSource
import net.minecraft.util.Mth
import net.minecraft.world.entity.Entity
import net.minecraft.world.entity.player.Player
import net.minecraft.world.item.ItemStack
import net.minecraft.world.phys.Vec3
import ru.dbotthepony.mc.otm.core.*
import ru.dbotthepony.mc.otm.core.Vector
import ru.dbotthepony.mc.otm.entity.PlasmaProjectile
import ru.dbotthepony.mc.otm.position
import ru.dbotthepony.mc.otm.registry.MSoundEvents
import java.util.*
import kotlin.math.PI
import kotlin.math.cos
import kotlin.math.sin

private val random = Random()
private const val DEGS = PI / 180.0

class VelocityCalculation(
	xRot: Double,
	yRot: Double,
	zRot: Double = 0.0,
	force: Double,
	deviation: Double = 0.0,
	owner: Entity? = null,
) {
	constructor(
		xRot: Float,
		yRot: Float,
		zRot: Float = 0f,
		force: Double,
		deviation: Double = 0.0,
		owner: Entity? = null,
	) : this(xRot.toDouble(), yRot.toDouble(), zRot.toDouble(), force, deviation, owner)

	constructor(
		xRot: Float,
		yRot: Float,
		zRot: Float = 0f,
		force: Float,
		deviation: Double = 0.0,
		owner: Entity? = null,
	) : this(xRot.toDouble(), yRot.toDouble(), zRot.toDouble(), force.toDouble(), deviation, owner)

	constructor(
		xRot: Float,
		yRot: Float,
		zRot: Float = 0f,
		force: Float,
		deviation: Float,
		owner: Entity? = null,
	) : this(xRot.toDouble(), yRot.toDouble(), zRot.toDouble(), force.toDouble(), deviation.toDouble(), owner)

	constructor(
		owner: Entity,
		force: Float,
		deviation: Float,
	) : this(owner.xRot.toDouble() - 4.5 * (1 / force), owner.yRot.toDouble(), 0.0, force.toDouble(), deviation.toDouble(), owner)

	constructor(
		owner: Entity,
		force: Double,
		deviation: Float,
	) : this(owner.xRot.toDouble() - 4.5 * (1 / force), owner.yRot.toDouble(), 0.0, force, deviation.toDouble(), owner)

	constructor(
		owner: Entity,
		force: Double,
		deviation: Double,
	) : this(owner.xRot.toDouble() - 4.5 * (1 / force), owner.yRot.toDouble(), 0.0, force, deviation, owner)

	val xVelocity = -sin(yRot * DEGS) * cos(xRot * (PI / 180.0))
	val yVelocity = -sin((xRot + zRot) * DEGS)
	val zVelocity = cos(yRot * DEGS) * cos(xRot * DEGS)

	val velocity: Vec3

	init {
		var velocity = Vec3(
			xVelocity,
			yVelocity,
			zVelocity
		).normalize()

		velocity += Vec3(
			random.nextGaussian() * 0.0075 * deviation,
			random.nextGaussian() * 0.0075 * deviation,
			random.nextGaussian() * 0.0075 * deviation,
		)

		velocity *= force

		if (owner != null) {
			val ownerVel = owner.deltaMovement

			if (owner.isOnGround) {
				velocity += ownerVel
			} else {
				velocity += Vec3(ownerVel.x, 0.0, ownerVel.z)
			}
		}

		this.velocity = velocity
	}

	val yRotation = Mth.atan2(velocity.x, velocity.z).toFloat() * (180f / Math.PI.toFloat())
	val xRotation = Mth.atan2(velocity.y, velocity.horizontalDistance()).toFloat() * (180f / Math.PI.toFloat())

	fun load(onto: Entity, withRotation: Boolean = true) {
		if (withRotation) {
			onto.xRot = xRotation
			onto.yRot = yRotation
		}

		onto.deltaMovement = velocity
	}
}

class PlasmaRifleItem : PlasmaWeaponItem<WeaponDataTable>(WeaponDataTable::class, ImpreciseFraction(200_000)) {
	override val roundsPerMinute: Int = 200
	override val scopingTime: Int = 7

	override val positionIdle: Vector
		get() = Vector(0.2,-0.4,-0.6)
	override val positionIronSights: Vector
		get() = Vector(0.0,-0.24,-1.0)
	override val rotIdle: Angle
		get() = Angle(0.0, -0.02, 0.0)
	override val rotIronSights: Angle
		get() = Angle(-0.02, 0.0, 0.0)

	override fun primaryFire(itemStack: ItemStack, player: Player, dt: WeaponDataTable): Boolean {
		if (!player.level.isClientSide) {
			val arrow = PlasmaProjectile(player.level)
			arrow.position = player.eyePosition

			val calc = VelocityCalculation(player, force = 4.0, deviation = 0.3)
			calc.load(arrow)

			arrow.owner = player

			player.level.addFreshEntity(arrow)
		} else {
			dt.doFireAnim(deviation = rotFireAnimDeviation)
		}

		player.level.playSound(
			player,
			player,
			MSoundEvents.RIFLE_SHOT,
			SoundSource.NEUTRAL,
			1f,
			1f
		)

		return true
	}

	companion object {
		private val ENERGY_PER_SHOT = ImpreciseFraction(3_000)
	}
}
